from pydantic import BaseModel, Field, model_validator

class SurQLEvent(BaseModel):
    """
        A pydantic SurQL event definition
        whenSDL accepts a list of strings (will be splitted by OR) or a single line string
        querySDL accepts a list of strings to allow multiline queries (can also be a single multiline string)
        TODO: check name is not a reserved keyword
    """
    name: str = Field(min_length=1)
    whenSDL: str | list[str] = Field(default="", description="event when definition, if list default to OR")
    querySDL: str | list[str] = Field(default="", description="event THEN definition")

    @model_validator(mode='after')
    def validate_event(self):
        assert len(self.querySDL) > 0, "event %s querySDL definition is required" % self.name
        assert len(self.whenSDL) > 0, "event %s WHEN close required" % self.name
        return self

    def SDL(self, table_name: str):
        """
            return a SDL event definition
        """
        _def = [
            "DEFINE EVENT",
            self.name,
            "ON TABLE",
            table_name,
            "WHEN",
        ]
        if (type(self.whenSDL) == str):
            _def += [self.whenSDL, "THEN"]
        else:
            _def += [" OR ".join(self.whenSDL), "THEN"]
        if (type(self.querySDL) == str):
            _def += ["(" + self.querySDL + ")"]
        else:
            _def += ["\n".join(["{", *self.querySDL, "}"])]
        return " ".join(_def) + ';'


from enum import Enum
from typing import LiteralString
from pydantic import BaseModel, Field, field_validator
import re

class SurQLTokenizers(Enum):
    """
        SurQL tokenizers enumeration
    """
    BLANK = "blank"
    CAMEL = "camel"
    CLASS = "class"
    PUNCT = "punct"

SNOWBALL_LANG: LiteralString = "|".join([
    "arabic",
    "danish",
    "dutch",
    "english",
    "french",
    "german",
    "greek",
    "hungarian",
    "italian",
    "norwegian",
    "portuguese",
    "romanian",
    "russian",
    "spanish",
    "swedish",
    "tamil",
    "turkish",
])
RE = re.compile(fr"^snowball\(({SNOWBALL_LANG})\)|ascii|lowercase|uppercase|edgengram\(\d+,\d+\)$")

class SurQLAnalyzer(BaseModel):
    """
        A pydantic SurQL analyzer definition
        TODO: check name is not a reserved keyword
    """
    name: str = Field(min_length=1)
    tokenizers: list[SurQLTokenizers] = Field(min_length=1)
    filters: list[str] = []

    @field_validator("filters")
    @classmethod
    def filters_validator(cls, v):
        """
            validate filters
        """
        assert len(v) == len(set(v)), "filters must be unique"
        for t in v:
            assert re.match(RE, t), "invalid filter, must be one of ascii, lowercase, uppercase, edgengram(<min>,<max>), snowball(<lang>)"
        return v

    def SDL(self) -> str:
        """
            return a SDL analyzer definition
        """
        _def = [
            "DEFINE ANALYZER",
            self.name,
            "TOKENIZERS",
            ','.join([t.value for t in self.tokenizers]),
        ]
        if (len(self.filters) > 0):
            _def += ["FILTERS", ','.join(self.filters)]
        return " ".join(_def) + ';'

class SurQLIndex(BaseModel):
    """
        A pydantic SurQL index definition
        TODO: check name is not a reserved keyword
    """
    name: str = Field(min_length=1)
    fields: list[str]

    def baseSDL(self, table_name: str) -> list[str]:
        """
            return a SDL index definition list of terms
        """
        return [
            "DEFINE INDEX",
            self.name,
            "ON TABLE",
            table_name,
            "FIELDS",
            f"{','.join(self.fields)}",
        ]

    def SDL(self, table_name: str) -> str:
        """
            return a SDL index definition
        """
        return " ".join([e for e in self.baseSDL(table_name) if e is not None]) + ';'

class SurQLUniqueIndex(SurQLIndex):
    """
        A pydantic SurQL unique index definition
    """
    def baseSDL(self, table_name: str) -> list[str]:
        """
            return a SDL unique index definition list of terms
        """
        return super().baseSDL(table_name) + ["UNIQUE"]

class SurQLSearchIndex(SurQLIndex):
    """
        A pydantic SurQL search index definition
    """
    analyzer: SurQLAnalyzer
    bm25: tuple[float, float] | None = None
    highlights: bool = False

    def baseSDL(self, table_name: str) -> list[str]:
        """
            return a SDL search index definition list of terms
        """
        return super().baseSDL(table_name) + [
            "SEARCH ANALYZER",
            self.analyzer.name,
            f"BM25({self.bm25[0]},{self.bm25[1]})" if self.bm25 else None,
            "HIGHLIGHTS" if self.highlights else None,
        ]


from enum import Enum
from pydantic import BaseModel
from datetime import datetime
from typing import Any, Optional, Type, Union, get_origin, get_args
from types import UnionType, NoneType, GenericAlias

from pydantic_surql.types.field import SurQLFieldInfo

from .cache import Cache
from .types import RecursiveType, SurQLAnyRecord, SurQLField, SurQLType, SurQLNullable, SurQLTable, SurQLTableConfig

def is_union(annotation: Type) -> bool:
    """
        Check if a type is a Union
    """
    origin = get_origin(annotation)
    return (origin == Union or origin == UnionType)

class SurQLParser:
    """
        A pydantic SurQL parser
    """
    def __init__(self):
        self.cache = Cache()

    @staticmethod
    def to_simple_type(_type: Type) -> SurQLType | None:
        """
            Parse simple type definitions
        """
        if (_type == str):
            return SurQLType.STRING
        if (_type == int or _type == float):
            return SurQLType.NUMBER
        if (_type == datetime.date or _type == datetime):
            return SurQLType.DATE
        if (_type == bool):
            return SurQLType.BOOLEAN
        if (_type == Any):
            return SurQLType.ANY
        if (_type == SurQLNullable):
            return SurQLType.NULL
        if (_type == NoneType):
            return SurQLType.OPTIONAL
        if (_type == SurQLAnyRecord):
            return SurQLType.ANY_RECORD
        return None

    def from_type(self, _type: Type) -> RecursiveType | SurQLField:
        """
            Parse a type:
            TODO: check for objects recursive types (will raise: must be converted to record)
        """
        if (self.cache.has(_type)):
            return self.cache.get(_type)

        # is a simple type
        simpleType = self.to_simple_type(_type)
        if (simpleType is not None):
            return self.cache.set(_type, simpleType)

        # is a list
        if (isinstance(_type, GenericAlias) and _type.__origin__ == list):
            args = get_args(_type)
            res = []
            for arg in args:
                simpleType = self.to_simple_type(arg)
                if (simpleType is not None):
                    res.append(simpleType)
                elif (is_union(arg)):
                    res += self.from_union(arg)
                else:
                    res.append(self.from_type(arg))
            return self.cache.set(_type, res)

        # is a set
        if (isinstance(_type, GenericAlias) and _type.__origin__ == set):
            args = get_args(_type)
            res = []
            for arg in args:
                simpleType = self.to_simple_type(arg)
                if (simpleType is not None):
                    res.append(simpleType)
                elif (is_union(arg)):
                    res += self.from_union(arg)
                else:
                    res.append(self.from_type(arg))
            return self.cache.set(_type, [SurQLType.SET, res])

        # is an enum
        if issubclass(_type, Enum):
            assertions = ",".join([f'"{item.value}"' if isinstance(item.value, str) else str(item.value) for item in _type])
            return SurQLField(name=None, types=[SurQLType.ENUM], assertion=f"$value in [{assertions}]")
        # is a pydantic model
        if (issubclass(_type, BaseModel)):
            if hasattr(_type, '__is_surql_collection__'):
                # was decorated with @surql_collection
                return SurQLField(name=None, types=[SurQLType.RECORD], recordLink=_type.__surql_table_name__)
            extra = _type.model_config.get('extra')
            if extra == 'allow':
                return SurQLField(name=None, types=self.from_fields(_type), recordLink=None, isFlexible=True)
            return SurQLField(name=None, types=self.from_fields(_type), recordLink=None)
        raise Exception(f"Type {_type} is not supported")

    def from_union(self, type: UnionType) -> RecursiveType:
        """
            Parse a Union type
        """
        types = []
        if (self.cache.has(type)):
            types = self.cache.get(type)
        else:
            for e in get_args(type):
                types.append(self.from_type(e))
        return self.cache.set(type, types)

    def from_field_type(self, annotation: Type) -> RecursiveType:
        """
            Parse a pydantic model field type to a SurQLField
        """
        if (is_union(annotation)):
            types = self.from_union(annotation)
        else:
            types = [self.from_type(annotation)]
        return types

    def from_field(self, name: Optional[str], annotation: Type) -> SurQLField:
        """
            Parse a pydantic model field to a SurQLField
        """
        types = self.from_field_type(annotation)
        return SurQLField(name=name, types=types)

    def from_fields(self, model: BaseModel) -> list[SurQLField]:
        """
            Parse a pydantic model to a list of SurQLField
        """
        fields = []
        for field_name, field in model.model_fields.items():
            perms = None
            if isinstance(field, SurQLFieldInfo):
                perms = field.perms
            is_collection = getattr(model, '__is_surql_collection__', False)
            if (field_name != 'id' or is_collection == False):
                _field = None
                _field = self.from_field(field_name, field.annotation)
                _field.perms = perms
                fields.append(_field)
        return fields

    def from_model(self, name: str, model: BaseModel, config: SurQLTableConfig = SurQLTableConfig()) -> SurQLTable:
        """
            Convert a pydantic model to a SurQLTable
            can be used at runtime
        """
        model.__is_surql_collection__ = True
        model.__surql_table_name__ = name
        extra = model.model_config.get('extra')
        if extra == 'allow':
            config.strict = False
        elif config.strict == False:
            model.model_config['extra'] = 'allow'
        return SurQLTable(name=name, fields=self.from_fields(model), config=config)


from typing import TypedDict, Type
from .types import RecursiveType, SurQLField

class Cache():
    """
        A simple cache system maping a Type to a list[RecursiveType]
    """
    def __init__(self):
        """
            Initialize the cache
        """
        self.cache: TypedDict[Type, RecursiveType | SurQLField] = {}

    def get(self, key: Type):
        """
            Get a key from the cache
        """
        return self.cache.get(key)

    def set(self, key: Type, value: TypedDict):
        """
            Set a key in the cache
        """
        self.cache[key] = value
        return value

    def has(self, key: Type):
        """
            Check if the cache has a key
        """
        return key in self.cache

    def clear(self):
        """
            Clear the cache
        """
        self.cache = {}


from pydantic import BaseModel, Field, field_validator

from .event import SurQLEvent
from .field import SurQLField
from .indexes import SurQLAnalyzer, SurQLIndex
from .permissions import SurQLPermissions

class SurQLView(BaseModel):
    """
        A pydantic SurQL view query definition
    """
    select: list[str] = Field(default=["*"], min_length=1)
    from_t: list[str] = Field(min_length=1)
    where: list[str]
    group_by: list[str]

    def SDL(self) -> str:
        """
            return a SDL view definition
        """
        _def = [
            "AS SELECT",
            ','.join(self.select),
            "FROM",
            ','.join(self.from_t),
        ]
        if (len(self.where) > 0):
            _def += ["WHERE", ','.join(self.where)]
        if (len(self.group_by) > 0):
            _def += ["GROUP BY", ','.join(self.group_by)]
        return " ".join(_def)

class SurQLTableConfig(BaseModel):
    """
        A pydantic SurQL table configuration definition
        TODO: add validation for changefeed
        TODO: implement table permissions
        TODO: implement table events
    """
    asView: SurQLView | None = Field(default=None, description="view definition")
    strict: bool = Field(default=True, description="schemafull|schemaless")
    changeFeed: str | None = Field(default=None, description="changefeed definition")
    drop: bool = Field(default=False, description="set table in DROP mode")
    indexes: list[SurQLIndex] = Field(default=[], description="table indexes definitions")
    events: list[SurQLEvent] = Field(default=[], description="table events definitions")
    permissions: SurQLPermissions | None = Field(default=None, description="table permissions definitions")

    @field_validator("indexes")
    @classmethod
    def indexes_validator(cls, v):
        """
            validate indexes
        """
        assert len(v) == len(set([index.name for index in v])), "indexes names must be unique"
        return v


class SurQLTable(BaseModel):
    """
        A pydantic SurQL table definition
        TODO: check name is not a reserved keyword
    """
    name: str = Field(min_length=1)
    fields: list[SurQLField]
    config: SurQLTableConfig = SurQLTableConfig()

    def _table_def(self):
        """return a SDL schemafull table definition"""
        _def = [
            "DEFINE TABLE",
            self.name,
        ]
        if (self.config.asView is not None):
            _def += [self.config.asView.SDL()]
        else:
            _def += [
                "DROP" if self.config.drop else None,
                "SCHEMAFULL" if self.config.strict else "SCHEMALESS",
                f"CHANGEFEED {self.config.changeFeed}" if self.config.changeFeed is not None else None,
            ]
            _def = [e for e in _def if e is not None]
        if (self.config.permissions is not None):
            _def.append(self.config.permissions.SDL())
        return " ".join(_def) + ';'

    def SDL(self):
        """return a SDL table definition with all the fields SDL definitions"""
        res = [self._table_def()]
        if (self.config.asView is None):
            for field in self.fields:
                res.append(field.SDL(self.name))
            for index in self.config.indexes:
                res.append(index.SDL(self.name))
            for event in self.config.events:
                res.append(event.SDL(self.name))
        return "\n".join(res)

class SurQLMetadata(BaseModel):
    """
        A simple mapper to store all the SurQL tables definitions generated from pydantic models through the decorator @surql_collection
    """
    tables: list[SurQLTable] = []
    analyzers: list[SurQLAnalyzer] = []

    @field_validator("tables")
    @classmethod
    def tables_validator(cls, v):
        """
            validate tables
        """
        assert len(v) == len(set([table.name for table in v])), "tables names must be unique"
        return v

    def clear(self):
        """
            clear all the tables definitions
        """
        self.tables = []
        self.analyzers = []

    def collect(self):
        """
            return a SDL string with all the tables definitions
        """
        res = []
        for analyzer in self.analyzers:
            res.append(analyzer.SDL())
        for table in self.tables:
            res.append(table.SDL())
        return "\n\n".join(res)


from pydantic import BaseModel, model_validator

class SurQLPermissions(BaseModel):
    """
        A pydantic SurQL table / field permissions definition
    """
    select: list[str] | None = None
    create: list[str] | None = None
    update: list[str] | None = None
    delete: list[str] | None = None

    @staticmethod
    def __term_sdl__(term: str, definitions: list[str]) -> str:
        """
            return a SDL permission block
        """
        _def = '\n'.join([f"        {e}" for e in definitions])
        return f"FOR {term}\n{_def}"

    def SDL(self):
        _def = [
            "PERMISSIONS",
            self.__term_sdl__("SELECT", self.select) if self.select is not None else None,
            self.__term_sdl__("CREATE", self.create) if self.create is not None else None,
            self.__term_sdl__("UPDATE", self.update) if self.update is not None else None,
            self.__term_sdl__("DELETE", self.delete) if self.delete is not None else None,
        ]
        return "\n    ".join([e for e in _def if e is not None])

    @model_validator(mode="after")
    @classmethod
    def check_permissions(cls, v):
        """
            check if permissions are valid
        """
        assert v.select is None or len(v.select) > 0, "SELECT permissions must have at least one permission set"
        assert v.create is None or len(v.create) > 0, "CREATE permissions must have at least one permission set"
        assert v.update is None or len(v.update) > 0, "UPDATE permissions must have at least one permission set"
        assert v.delete is None or len(v.delete) > 0, "DELETE permissions must have at least one permission set"
        assert any([v.select != None, v.create != None, v.update != None, v.delete != None]), "Permissions must have at least one permission set"
        return v


from typing import List, Type, Union, Sequence
from enum import Enum
from typing import Optional
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo
from typing_extensions import TypeAliasType

from .permissions import SurQLPermissions

"""
    A custom type to define a nullable field
"""
SurQLNullable =  Type[None]

"""
    A custom type to define a generic record
"""
SurQLAnyRecord = Type[dict]

class SurQLType(Enum):
    """
        SurQL types enumeration
    """
    STRING = "string"
    NUMBER = "number"
    DATE = "datetime"
    ANY = "any"
    BOOLEAN = "bool"
    ENUM = "enum"
    ARRAY = "array"
    OBJECT = "object"
    SET = "set"
    RECORD = "record<%s>"
    ANY_RECORD = "record()"
    OPTIONAL = "optional<%s>"
    NULL = "null"

BASIC_TYPES: list[SurQLType] = [
    SurQLType.STRING,
    SurQLType.NUMBER,
    SurQLType.ANY_RECORD,
    SurQLType.DATE,
    SurQLType.BOOLEAN,
    SurQLType.ANY,
    SurQLType.NULL,
]

COMPLEX_TYPES = [
    SurQLType.OBJECT,
    SurQLType.RECORD,
]

RecursiveType = TypeAliasType('RecursiveType', Sequence[Union[SurQLType, 'SurQLField', 'RecursiveType']])

class SurQLField(BaseModel):
    """
        A pydantic SurQL field definition
        TODO: implement permissions
    """
    name: Optional[str]
    types: RecursiveType
    recordLink: Optional[str] = None
    isFlexible: bool = False
    perms: Optional[SurQLPermissions] = None
    assertion: Optional[str] = None

    @classmethod
    def _f_string(
        cls,
        field: str,
        table: str,
        types: str,
        isFlexible: bool,
        assertions: Optional[list[str]] = None,
        perms: Optional[SurQLPermissions] = None
    ):
        """
            return a SDL field definition string
        """
        return " ".join(e for e in [
            f"DEFINE FIELD {field} ON TABLE {table} {'FLEXIBLE ' if isFlexible else ''}TYPE {types}",
            f"ASSERT ({' OR '.join(assertions)})" if assertions is not None and len(assertions) > 0 else None,
            perms.SDL() if perms is not None else None
        ] if e != None) + ";"

    @classmethod
    def _surqlFromTypes(cls, table_name: str, field_name: str, types: List[Type], perms: Optional[SurQLPermissions] = None) -> list[str]:
        """
            return SDLS fields definitions recursively
            TODO: remove duplicates (eg: when a field is defined as int | float)
            TODO: check name is not a reserved keyword
        """
        res = []
        nextFields = []
        assertions = []
        isOptional = False
        isFlexible = False
        for _type in types:
            if (_type in BASIC_TYPES):
                res += [_type.value]
            # if (_type is SurQLType.ENUM):
            #     res += []
            elif isinstance(_type, list):
                if (_type[0] == SurQLType.SET):
                    res += [SurQLType.SET.value]
                    nextFields += cls._surqlFromTypes(table_name, f"{field_name}.*", _type[1])
                else:
                    res += [SurQLType.ARRAY.value]
                    _perms = getattr(_type, "perms", None)
                    nextFields += cls._surqlFromTypes(table_name, f"{field_name}.*", _type, _perms)
            elif (isinstance(_type, cls)):
                if (_type.types == [SurQLType.RECORD]):
                    res += [SurQLType.RECORD.value % _type.recordLink]
                elif (_type.types == [SurQLType.ENUM]):
                    res += [SurQLType.STRING.value, SurQLType.NUMBER.value]
                    assertions += [_type.assertion]
                else:
                    res += [SurQLType.OBJECT.value]
                    isFlexible = _type.isFlexible
                    for _field in _type.types:
                        _perms = getattr(_field, "perms", None)
                        nextFields += cls._surqlFromTypes(table_name, f"{field_name}.{_field.name}", _field.types, _perms)
            elif (_type is SurQLType.OPTIONAL):
                isOptional = True
            else:
                raise Exception(f"Unknown type: {_type}, SDL generation not supported")
        if (isOptional):
            return [cls._f_string(field_name, table_name, SurQLType.OPTIONAL.value % "|".join(res), isFlexible, assertions, perms)] + nextFields
        return [cls._f_string(field_name, table_name, "|".join(res), isFlexible, assertions, perms)] + nextFields


    def SDL(self, table_name: str) -> List[str]:
        """return a SDL field definition"""
        fieldTypes = SurQLField._surqlFromTypes(table_name, self.name, self.types, self.perms)
        return "\n".join(fieldTypes)


    __hash__ = object.__hash__

SurQLField.model_rebuild()

class SurQLFieldInfo(FieldInfo):
    """
        A pydantic SurQL field info definition
    """
    perms: Optional[SurQLPermissions] = None

    def __init__(self, perms: Optional[SurQLPermissions], **kwargs):
        super().__init__(**kwargs)
        self.perms = perms

def SurQLFieldConfig(permissions: Optional[SurQLPermissions] = None, **kwargs) -> SurQLFieldInfo:
    """
        A pydantic SurQL field config definition
        TODO: find a way to map Field arguments for proper type hints (see: https://stackoverflow.com/questions/1409295/set-function-signature-in-python)
    """
    return SurQLFieldInfo(permissions, **kwargs)


from pydantic import BaseModel
from .parser import SurQLParser
from .types import SurQLMetadata, SurQLTableConfig

Parser = SurQLParser()
Metadata = SurQLMetadata()

def surql_collection(name: str, config: SurQLTableConfig = SurQLTableConfig()):
    """
        A simple decorator to convert a pydantic model to a surQL SDL table definition
    """
    def inner(model: BaseModel):
        table = Parser.from_model(name, model, config)
        Metadata.tables += [table]
        for index in table.config.indexes:
            if hasattr(index, "analyzer") and index.analyzer is not None:
                exist = any([a.name == index.analyzer.name for a in Metadata.analyzers])
                if not exist:
                    Metadata.analyzers += [index.analyzer]
        return model
    return inner